#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xresource.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>
#include <X11/xpm.h>

#include "othello.h"

static int player_id[2] = {0, 1};

////////////////////////////////////////////////////////////////
// Othello routines
////////////////////////////////////////////////////////////////

long think_t[2];
participant_t *think_fun[2];
static char *player_name[2] = {"", ""};
char m_board[64], v_board[64], board[64];
int hands;
int turn;
int last_hand;

const int direction_x[] = {-1,  0,  1, -1,    1, -1,  0,  1};   /* 0 1 2 */
const int direction_y[] = {-1, -1, -1,  0,    0,  1,  1,  1};   /* 3 @ 4 */
const int direction[] =   {-9, -8, -7, -1,    1,  7,  8,  9};   /* 5 6 7 */

static void InitBoard(char *m_board)
{
  turn = 0;
  hands = 1;

  int i;

  /* Set Initial State */
  for(i = 0; i < 64; i++)
    m_board[i] = SPACE;

  m_board[27] = m_board[36] = WHITE; /*  */
  m_board[28] = m_board[35] = BLACK; /*  */
}

void initiate_game_1()
{
  think_fun[0] = e_list[player_id[0]].func;
  think_fun[1] = e_list[player_id[1]].func;
  think_t[0] = think_t[1] = 0;
  InitBoard(m_board);
  last_hand = -1;
}

void initiate_game_2()
{
  char board[64];
  CopyBoard(m_board, board, NORM);
  if(think_fun[0])
    think_fun[0](board, INIT_AS_BLACK);

  CopyBoard(m_board, board, REVERSE);
  if(think_fun[1])
    think_fun[1](board, INIT_AS_BLACK);
}

static void RepairBoard(char *board)
{
  int i;

  for(i = 0; i < 64; i++)
    if(board[i] < 0)
      board[i] = -board[i];
}


////////////////////////////////////////////////////////////////
// X routines
////////////////////////////////////////////////////////////////

#define XAW3D_NOSHADOW "shadowWidth", 0, XtNborderWidth, 0, 

static Display *X_d = 0;
static Window X_root = 0;
static XtAppContext X_app = NULL;
static Atom XA_wm_protocols, XA_wm_delete_window;

#define CELLSIZE 36
#define CELLMARGIN 5

static String X_fallbacks[] = {
// app resources
//  "interactive: False",
//  "*fullLeague: False",
//  "*silent: False",
  "*help: False",

// global
  "*Label.justify: left",
  "*Label.resize: False",
  "*Command.resize: False",
  "*showGrip: off",

// title
  "*game*title.Label: Othello X GUI version 0.1.0",
  "*turnLabel.Label: Turn",
  "*turn.justify: right",
  "*turn.width: 30",
  "*titleBar.showGrip: off",

// scores
  "*game*label.width: 50",
  "*game*name.width: 150", 
  "*game*score.width: 40",
  "*game*score.justify: right",

  "*game*black.borderWidth: 0",
  "*game*black.label.label: Black:",
  "*game*white.label.label: White:",

// board
  "*othelloBoard.background: darkgreen",
  "*othelloBoard.width: 340",
  "*othelloBoard.height: 340",
  "*cell.background: darkgreen",
  "*cell.internalWidth: 2",
  "*cell.internalHeight: 2",
  "*cell.foreground: white",

// play bar
  "*pass.label: Pass",

// text region

// command bar
  "*start.label: Start",
  "*quit.label: Quit",

// popup
  "*select*Label.width: 200",
  "*select*Label.height: 20",
  "*select*black.borderWidth: 0",
  "*select*white.borderWidth: 0",
  "*select*black*label.label: Black Player:",
  "*select*white*label.label: White Player:",
  "*select*white.fromHoriz: black",
  "*select*List.lines: 20",
  "*select*view.fromVert: label",
  "*select*view.allowVert: true",
  "*select*List.width: 200",
  "*select*Viewport.width: 200",
  "*select*Viewport.height: 200",
//  "*select*bar.fromVert: black",
  "*select*ok.fromVert: black",
  "*select*ok.label: OK",
  "*select*cancel.label: Cancel",
  "*select*cancel.fromHoriz: ok",
  "*select*cancel.fromVert: black",

  NULL
};

// X variables

static enum {
  COMPTURN, HUMANTURN, HUMANPASS, THINKING, NOTSTARTED
} X_status = NOTSTARTED;

static Widget XW_shell, XW_form, XW_board, XW_cell[64], XW_message;
static Widget XW_score_board[2], XW_score_name[2], XW_score_num[2];
static Widget XW_nav_bar, XW_pass_btn;
static Widget XW_command_bar, XW_start_btn, XW_quit_btn;
static Widget XW_titlebar, XW_turn_num;
static Pixmap pix_empty, pix_black[4], pix_white[4];

static XrmOptionDescRec opts_parse[] = {
  { "-interactive", ".interactive", XrmoptionNoArg, (XPointer)True },
  { "-silent", ".silent", XrmoptionNoArg, (XPointer)True },
  { "-fullleague", ".fullLeague", XrmoptionNoArg, (XPointer)True },
  { "-help", ".help", XrmoptionNoArg, (XPointer)True },
  { "--help", ".help", XrmoptionNoArg, (XPointer)True },
  { "-?", ".help", XrmoptionNoArg, (XPointer)True }
};

static String X_shell_translation = "<ClientMessage>: destroy()";

static Widget XpW_popup, XpW_frame, XpW_f0, XpW_f1, XpW_label0, XpW_label1, XpW_list0, XpW_list1;
static Widget XpW_okbtn, XpW_cancelbtn;

static char *X_Colorname[] = {"Black", "White"};
static char *X_colorname[] = {"black", "white"};

////////////////////////////////////////////////////////////////
// Callbacks
////////////////////////////////////////////////////////////////

static void X_message(char *format, ...)
{
  va_list va;
  va_start(va, format);
  static char buf[256];
  vsprintf(buf, format, va);

  XtVaSetValues(XW_message,
		XtNlabel, buf,
		NULL);
 
  va_end(va);
}

static Pixmap X_currentpic[63];

static void X_setboard(char *board, int lasthand)
{
  for(int i = 0; i <= 63; i++) {
    int p = board[i];
    Pixmap New;
    int o = (i == lasthand) ? 3 : (p < 0 ? 2 : 0);

    if (p == 0) 
      New = pix_empty;
    else if (p == BLACK || p == -BLACK)
      New = pix_black[o];
    else
      New = pix_white[o];

    if (X_currentpic[i] != New) {
      X_currentpic[i] = New;
      XtVaSetValues(XW_cell[i], XtNbitmap, New, NULL);
    }
  }
}

static void X_update_state()
{
  char buf[32];

  X_setboard(v_board, last_hand);
  sprintf(buf, "%d", hands);
  XtVaSetValues(XW_turn_num, XtNlabel, buf, NULL);

  int b, w;

  CountDisks(m_board, &b, &w);

  sprintf(buf, "%d", b);
  XtVaSetValues(XW_score_num[0], XtNlabel, buf, NULL);
  sprintf(buf, "%d", w);
  XtVaSetValues(XW_score_num[1], XtNlabel, buf, NULL);
}  

inline static void X_SetButtonActive(Widget w, Bool f)
{
  XtVaSetValues(w, XtNsensitive, f, NULL);
}

static bool isEndgame(void)
{
  for(int i = 0; i <= 63; i++) {
    if (Put(m_board, i, BLACK, NOT_CHANGE)
	|| Put(m_board, i, WHITE, NOT_CHANGE))
      return false;
  }
  return true;
}

static bool isPass(char *board, int c)
{
  for(int i = 0; i <= 63; i++) {
    if (Put(board, i, c, NOT_CHANGE))
      return false;
  }
  return true;
}

static void X_setstate_startwait()
{
  for(int i = 0; i <= 63; i++) {
    X_SetButtonActive(XW_cell[i], False);
  }
  X_SetButtonActive(XW_pass_btn, False);
  X_SetButtonActive(XW_start_btn, True);
}

static void X_setup_for_next();

static void X_put_chip(int i)
{
  Put(m_board, i, turn + 1, R_CHANGE);
  CopyBoard(m_board, v_board, NORM);
  RepairBoard(m_board);

  last_hand = i;
  turn ^= 1;

  X_update_state();
  
  X_setup_for_next();
}

static void X_pass_chip()
{
  turn ^= 1;
  last_hand = -1;
  
  RepairBoard(v_board);
  
  X_update_state();
  X_setup_for_next();
}

static void X_comp_timer_event(XtPointer cdat, XtIntervalId *)
{
  if (X_status != COMPTURN
      || (int)cdat != hands * 2 + turn)
    return;

  CopyBoard(m_board, board, (turn ? REVERSE : NORM));

  if (isPass(board, MY_COLOR)) {
    think_fun[turn](board, PASS);
    X_pass_chip();
  } else {
    int p = think_fun[turn](board, NORM);
  
    if (p < 0 || p > 63 || Put(board, p, MY_COLOR, NOT_CHANGE) < 1) {
      printf("invalid ret: %d\n", p);
      exit(1);
    }

    X_put_chip(p);
  }
}

static void X_setup_for_next()
{
  if (isEndgame()) {
    X_update_state();

    int b, w;
    CountDisks(m_board, &b, &w);

    if (b > w)
      X_message("Game Finished.  White (%s) lose. (diff=%d)", player_name[1], b - w);
    else if (b < w)
      X_message("Game Finished.  Black (%s) lose. (diff=%d)", player_name[0], w - b);
    else
      X_message("Game Finished.  Draw. (diff=0).");

    X_setstate_startwait();
    return;
  }
  if (last_hand >= 0)
    hands++;

  X_update_state();
  
  if (think_fun[turn]) {
    X_status = COMPTURN;
    X_message("%s (%s)'s turn.", X_Colorname[turn], player_name[turn]);
    XtAppAddTimeOut(X_app, 1000, X_comp_timer_event, (XtPointer)(hands * 2 + turn));
  } else {
    X_status = HUMANTURN;
    // Human
    
    int passp = 1;
    for(int i = 0; i <= 63; i++) {
      if (Put(m_board, i, turn + 1, NOT_CHANGE)) {
	passp = 0;
	X_SetButtonActive(XW_cell[i], True);
      } else {
	X_SetButtonActive(XW_cell[i], False);
      }
    }
    if (passp) {
      X_message("%s (human)'s turn. You have to Pass.", X_Colorname[turn]);
      X_status = HUMANPASS;
      X_SetButtonActive(XW_pass_btn, True);
    } else {
      X_message("%s (human)'s turn. Put your stone.", X_Colorname[turn]);
      X_SetButtonActive(XW_pass_btn, False);
    }
  }
}

static void XC_initiategame(Widget, XtPointer, XtPointer)
{
  XtPopdown(XpW_popup);

  player_name[0] = e_list[player_id[0]].name;
  player_name[1] = e_list[player_id[1]].name;

  XtVaSetValues(XW_score_name[0], 
		XtNlabel, player_name[0],
		NULL);
  XtVaSetValues(XW_score_name[1], 
		XtNlabel, player_name[1],
		NULL);

  initiate_game_1();

  CopyBoard(m_board, v_board, NORM);
  X_update_state();
  XFlush(X_d);

  initiate_game_2();
  
  X_setup_for_next();
}

static void XC_pushcell(Widget, XtPointer cdat, XtPointer)
{
  int i = (int)cdat;

  if (X_status != HUMANTURN ||
      !Put(m_board, i, turn + 1, NOT_CHANGE)) {
    XBell(X_d, 50);
    return;
  }
  X_put_chip(i);
}


static void XC_pass(Widget, XtPointer, XtPointer)
{
  if (X_status != HUMANPASS)
    return;

  X_pass_chip();
}


static void XC_selectplayer(Widget, XtPointer, XtPointer)
{
  XtPopup(XpW_popup, XtGrabNone);
}

static void XC_quitapp(Widget, XtPointer, XtPointer)
{
  exit(0);
}

static void XM_destroy_message(Widget, XEvent *e, String *, Cardinal *)
{
  if (e->type == ClientMessage &&
      e->xclient.message_type == XA_wm_protocols &&
      (unsigned)e->xclient.data.l[0] == XA_wm_delete_window) {
    exit(0);
  }
}

static void XC_canceldialog(Widget, XtPointer, XtPointer)
{
  XtPopdown(XpW_popup);
}

static void XC_userselected(Widget, XtPointer cdat, XtPointer rdat)
{
  int i = (int)cdat;
  XawListReturnStruct *lrs = (XawListReturnStruct *)rdat;

  assert(i < 2);

  player_id[i] = lrs->list_index;

//  printf("player_id[%d] = %d\n", i, lrs->list_index);
}

////////////////////////////////////////////////////////////////
// init functions
////////////////////////////////////////////////////////////////

static XtActionsRec X_actions[] =
{
  {"destroy", XM_destroy_message },
};

bool X_initialize_application(int *argcp, char **argv)
{
  if (XW_shell) 
    return True;

  XW_shell = XtVaOpenApplication(&X_app, 
				 "XOthello",
				 opts_parse, XtNumber(opts_parse), 
				 argcp, argv,
				 X_fallbacks,
				 applicationShellWidgetClass, 
				 NULL);

  if (!XW_shell)
    return false;

  X_d = XtDisplay(XW_shell);
  X_root = RootWindow(X_d, 0);

  XtAppAddActions(X_app, X_actions, XtNumber(X_actions));

  return true;
}

// DUMMY
extern "C" int human(char *, int)
{
  abort();
}

void X_create_selectwindow()
{
// Setup List
  static char *names[100];
  int nlist;
  for(nlist = 0; e_list[nlist].name; nlist++) {
    names[nlist] = e_list[nlist].name;
  }
  nlist--;

// WIDGETS
  XpW_popup = XtVaCreatePopupShell("select", transientShellWidgetClass, XW_shell,
				  NULL);

  XpW_frame = XtVaCreateManagedWidget("form", formWidgetClass, XpW_popup, NULL);
  XpW_f0 = XtVaCreateManagedWidget("black", formWidgetClass, XpW_frame, NULL);
  XpW_label0 = XtVaCreateManagedWidget("label", labelWidgetClass, XpW_f0, 
				     NULL);
  XpW_list0 = XtVaCreateManagedWidget("player", listWidgetClass,
				    XtVaCreateManagedWidget("view", viewportWidgetClass, XpW_f0, NULL),
				    XtNlist, names,
				    XtNnumberStrings, nlist,
				    NULL);
  XawListHighlight(XpW_list0, 0);
  XtAddCallback(XpW_list0, XtNcallback, XC_userselected, 0);

  XpW_f1 = XtVaCreateManagedWidget("white", formWidgetClass, XpW_frame, NULL);
  XpW_label1 = XtVaCreateManagedWidget("label", labelWidgetClass, XpW_f1,
				     NULL);
  XpW_list1 = XtVaCreateManagedWidget("player", listWidgetClass,
				    XtVaCreateManagedWidget("view", viewportWidgetClass, XpW_f1, NULL),
				    XtNlist, names,
				    XtNnumberStrings, nlist,
				    NULL);
  XawListHighlight(XpW_list1, 1);
  XtAddCallback(XpW_list1, XtNcallback, XC_userselected, (XtPointer)1);
  
  XpW_okbtn = XtVaCreateManagedWidget("ok", commandWidgetClass, XpW_frame, //XpW_btnbar,
				    NULL);
  XtAddCallback(XpW_okbtn, XtNcallback, XC_initiategame, NULL);

  XpW_cancelbtn = XtVaCreateManagedWidget("cancel", commandWidgetClass, XpW_frame, //XpW_btnbar,
				      NULL);
  XtAddCallback(XpW_cancelbtn, XtNcallback, XC_canceldialog, NULL);

}

#include "pics/empty.xpm"
#include "pics/black.xpm"
#include "pics/black-gray.xpm"
#include "pics/black-plus.xpm"
#include "pics/black-star.xpm"
#include "pics/white.xpm"
#include "pics/white-gray.xpm"
#include "pics/white-plus.xpm"
#include "pics/white-star.xpm"

void X_create_gamewindow()
{
// Accept WM Delete
  XA_wm_protocols    =XInternAtom(X_d, "WM_PROTOCOLS",     False);
  XA_wm_delete_window=XInternAtom(X_d, "WM_DELETE_WINDOW", False);

  XtAugmentTranslations(XW_shell, 
			XtParseTranslationTable(X_shell_translation));

#define LOAD(from,to) XpmCreatePixmapFromData(X_d, X_root, from, to, NULL, NULL)

  LOAD(empty_xpm, &pix_empty);
  LOAD(black_xpm, &pix_black[0]);
  LOAD(black_gray_xpm, &pix_black[1]);
  LOAD(black_plus_xpm, &pix_black[2]);
  LOAD(black_star_xpm, &pix_black[3]);
  LOAD(white_xpm, &pix_white[0]);
  LOAD(white_gray_xpm, &pix_white[1]);
  LOAD(white_plus_xpm, &pix_white[2]);
  LOAD(white_star_xpm, &pix_white[3]);

#undef LOAD

  XW_form = XtVaCreateManagedWidget("game", panedWidgetClass, XW_shell, NULL);

// title bar
  XW_titlebar = XtVaCreateManagedWidget("titleBar", boxWidgetClass, XW_form, NULL);
  XtVaCreateManagedWidget("title", labelWidgetClass, XW_titlebar, NULL);
  XtVaCreateManagedWidget("turnLabel", labelWidgetClass, XW_titlebar,
			  XAW3D_NOSHADOW
			  NULL);
  XW_turn_num = XtVaCreateManagedWidget("turn", labelWidgetClass, XW_titlebar,
					XtNlabel, "0",
					NULL);

// score board
  for(int i=0; i < 2; i++) {
    static char *widnames[2] = { 
      "black", "white",
    };

    XW_score_board[i] = XtVaCreateManagedWidget(widnames[i], 
					     boxWidgetClass, XW_form, NULL);
    XtVaCreateManagedWidget("label", labelWidgetClass, 
			    XW_score_board[i], 
			    XAW3D_NOSHADOW
			    NULL);
    XW_score_name[i] = XtVaCreateManagedWidget("name", labelWidgetClass, 
					       XW_score_board[i], 
					       XAW3D_NOSHADOW
					       NULL);
    XW_score_num[i] = XtVaCreateManagedWidget("score", labelWidgetClass, 
					      XW_score_board[i], 
					      XtNlabel, "0",
					      NULL);
  }

// game board
  XW_board = XtVaCreateManagedWidget("othelloBoard", compositeWidgetClass, XW_form, NULL);

  for (int i = 0; i <= 63; i++) {
//    char buf[16], buf2[16];
//    sprintf(buf, "cell%02o", i);
//    sprintf(buf2, "%02o", i);
    int x = i % 8, y = i / 8;
    unsigned xgeom = x * (CELLSIZE + CELLMARGIN) + CELLMARGIN;
    unsigned ygeom = y * (CELLSIZE + CELLMARGIN) + CELLMARGIN;
    XW_cell[i] = XtVaCreateManagedWidget("cell", commandWidgetClass, XW_board, 
				      XtNbitmap, pix_empty,
				      XtNx, xgeom,
				      XtNy, ygeom,
				      NULL
				      );
    XtAddCallback(XW_cell[i], XtNcallback, XC_pushcell, (XtPointer)i);
  }

  XW_nav_bar = XtVaCreateManagedWidget("playBar", boxWidgetClass, XW_form, NULL);

  XW_pass_btn = XtVaCreateManagedWidget("pass", commandWidgetClass, XW_nav_bar, NULL);
  XtAddCallback(XW_pass_btn, XtNcallback, XC_pass, 0);

  XW_message = XtVaCreateManagedWidget("message", labelWidgetClass, XW_form, 
				       XtNlabel, "", NULL);

  XW_command_bar = XtVaCreateManagedWidget("commandBar", boxWidgetClass, XW_form, NULL);

  XW_start_btn = XtVaCreateManagedWidget("start", commandWidgetClass, XW_command_bar, NULL);
  XtAddCallback(XW_start_btn, XtNcallback, XC_selectplayer, 0);

  XW_quit_btn = XtVaCreateManagedWidget("quit", commandWidgetClass, XW_command_bar, NULL);
  XtAddCallback(XW_quit_btn, XtNcallback, XC_quitapp, 0);

  static char dummy_state[] = {
    0,0,0,0,0,0,0,0,
    1,1,1,2,2,0,1,1,
    0,1,2,0,0,1,0,0,
    0,1,0,2,0,1,0,1,
    0,1,0,0,2,1,0,1,
    0,1,2,2,0,0,1,1,
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
  };

  X_setboard(dummy_state, 034);

  X_setstate_startwait();

  XtRealizeWidget(XW_shell);
  XSetWMProtocols(X_d, XtWindow(XW_shell), &XA_wm_delete_window, 1);
}

int main(int argc, char **argv)
{
// for compatibility with othello.o
  assert (e_list[0].func == human);
  e_list[0].func = NULL;

  X_initialize_application(&argc, argv);
  X_create_gamewindow();

  X_create_selectwindow();

  XtAppMainLoop(X_app);
  return 0;
}

