/*
 *	taka's thinking routine
 *		for the Weakest Othello League
 *
 *	$Header: /home/taka/prog/oth/RCS/taka.cc,v 1.13 1997/10/10 04:23:39 taka Rel $
 */

#define TAKADEBUG 0
#define Thinker taka1Think1


#include <stdlib.h>
#include <string.h>
#if TAKADEBUG
#include <iostream.h>
#include <iomanip.h>
#endif

#include "othello.h"


/*----------------------------------------------------------------------*/
#if TAKADEBUG*0
struct _new_watcher {
	static long int count;
	_new_watcher() {;}
	~_new_watcher();
} _Anew_watcher;
long int _new_watcher::count=0;
_new_watcher::~_new_watcher()
{
	if (count) {
		cout <<"new/delete count = " <<count <<endl;
	}
}
void *operator new(size_t siz)
{
	_new_watcher::count++;
	return malloc(siz);
}
void operator delete(void *p)
{
	_new_watcher::count--;
	if (p!=NULL) free(p);
}
#endif	/* for TAKADEBUG */


/*----------------------------------------------------------------------*/
/*
 * constants and macros and type definitions
 */
#define TABLE_SIZE 91
#define tbl(x, y) ((y)*9 + (x))
#define tblx(pos) ((pos)%9)
#define tbly(pos) ((pos)/9)
#define tblresult(pos) (tbly(pos)*8 + tblx(pos) -9)

enum { P_WALL = -1, P_SPACE = 0, P_MY_COLOR = 1, P_OP_COLOR = 2 };
typedef signed char cint_t;
typedef cint_t piece_t;
typedef cint_t getlist_t;

/* for debug use */
#if TAKADEBUG
#define STATIC
#define DEADBEEF 0x34810567
#else
#define STATIC static
#endif


/*----------------------------------------------------------------------*/

struct taka1Malloc_block_header;

struct taka1Malloc_block_control {
	void *free;
	unsigned int block_size;
	unsigned int allocated_blocks;
	struct taka1Malloc_block_header *block_root;
};

struct taka1Malloc_block_header {
	struct taka1Malloc_block_header *next;
	unsigned int block_size;
	unsigned int allocated_blocks;
	struct taka1Malloc_block_control *pseudo_this;
};


STATIC void *Malloc_block_morecore(struct taka1Malloc_block_control *mbc)
{
	struct taka1Malloc_block_header *mbh;
	unsigned int l,s,*p;

	s=mbc->block_size;
	l=mbc->allocated_blocks;
	mbh=(struct taka1Malloc_block_header *)
	  malloc(sizeof(struct taka1Malloc_block_header)+s*l);
	if (!mbh) return NULL;
	mbh->next=mbc->block_root;
	mbh->block_size=s;
	mbh->allocated_blocks=l;
	mbh->pseudo_this=mbc;
	mbc->block_root=mbh;
	p=(unsigned int *)(mbh+1);
	while (--l) { *p=(unsigned int)p+s; p=(unsigned int *)((unsigned int)p+s); }
	*p=NULL;
	p=(unsigned int *)(mbh+1);
	mbc->free=(void *)*p;
	return p;
}

inline STATIC void *
Malloc_block(struct taka1Malloc_block_control *mbc)
{
	unsigned int *p;

	p=(unsigned int *)mbc->free;
	if (p) {
		mbc->free=(void *)*p;
		return p;
	}else {
		return Malloc_block_morecore(mbc);
	}
}

inline STATIC void
Free_block(struct taka1Malloc_block_control *mbc,void *block)
{
	*(unsigned int *)block=(unsigned int)mbc->free;
	mbc->free=block;
}

STATIC struct taka1Malloc_block_control Malloc256 =
        { 0, 256, 1024, 0 };


/*----------------------------------------------------------------------*/
/*
 * class definitions
 */
class taka1Board {
public:
	piece_t t[TABLE_SIZE];
	cint_t mycount;
	cint_t matchcount;
	cint_t totalcount;
	cint_t step;		/* my steps */
	cint_t myflag;		/* True if this object is from my own view */
private:
	static int rev_table[4];

public:
	taka1Board();
	taka1Board(taka1Board &src);
	taka1Board(char *initboard);
	int table(int pos);
	void settable(int pos, int d);
	void put_reverse(taka1Board &dstboard);	// for update
	taka1Board *create_reverse();
	
	// for debug use
#if TAKADEBUG
	void dump();
#endif
};

class Thinker;
class taka1Hand;
class taka1Choice;
class taka1Think {
	taka1Hand *handp;
public:	
	int weight;
protected:
	
	taka1Think(taka1Think &src);	// cannot call
	taka1Think();				// cannot call
	taka1Think(taka1Hand *handp);
	taka1Hand &hand();
	taka1Board &board();
public:
	~taka1Think();
	
	// standard interfaces
	//taka1Think(taka1Hand *handp);
	//static void initialize();
	//static int think(taka1Board &bd, taka1Choice &chfirst);
};

/*
 * Thinking routine #1: 1997/09/26(Fri.)
 */
class taka1Think1 : public taka1Think {
public:
	// standard interfaces
	taka1Think1(taka1Hand *handp);
	~taka1Think1();
	static void initialize();
	static int think(taka1Board &bd, taka1Choice &chfirst);
	
private:
	// local data and functions
	taka1Choice *cp;
	/* taka1Board *bp; */
	int aheadresult;
	static int weight_table_seed[4][4];
	static int weight_table_seed2[4][4];
	static int weight_table[TABLE_SIZE];
	
	void eval();
	int finalize(int passcount, int orgmin);
	int semifinalize(int passcount, int orgmin);
	int nthforward(int passcount, int orgmin, int depth);
};
/* ----- */

struct taka1Hand {
	enum { maxhand=22, passed=-1 };
	taka1Board bd;
	getlist_t getlist[maxhand];
	cint_t pos;
	cint_t getcount;
private:
	// variables for thinking...
	Thinker th;
	
public:
	void *operator new(size_t siz);
	void operator delete(void *p);
	taka1Hand();
	~taka1Hand();
	Thinker &think();
	
	// for debug use
	//void dump();
};

class taka1Choice {
	enum { maxhand=60 };
	taka1Hand *hv[maxhand];
	int hc;
	static taka1Hand *handstock;

protected:
	taka1Choice(taka1Choice &src);	// cannot call
	taka1Choice();					// cannot call
public:
	void *operator new(size_t siz);
	void operator delete(void *p);
	~taka1Choice();
	taka1Choice(taka1Board &from);
	void erasehand(int handcount);
	int number();
	taka1Hand &item(int hnum);	// const taka1Hand &
private:
	int makeput(taka1Hand *handp, taka1Board &from, int pos);
public:
	void handsort(int usecount = 9999);
	// for debug use
	//void dump();
};


/*-----*/
inline void *taka1Hand::operator new(size_t /*siz*/)
{ return Malloc_block(&Malloc256); }
inline void taka1Hand::operator delete(void *p)
{ Free_block(&Malloc256, p); }
inline void *taka1Choice::operator new(size_t /*siz*/)
{ return Malloc_block(&Malloc256); }
inline void taka1Choice::operator delete(void *p)
{ Free_block(&Malloc256, p); }


/*----------------------------------------------------------------------*/
/*
 * tables
 */
STATIC int intable_table[8*8];

static const int moves_table[8] = {
	-9-1, -9, -9+1,
	  -1,       +1,
	 9-1,  9,  9+1,
};


/*
 * global variables
 */
STATIC int my_step;


/*
 * prototype declarations
 */
static void initialize_tables(void);


/*----------------------------------------------------------------------*/
/*
 * class functions
 */

// class taka1Board
int taka1Board::rev_table[4] = { -1, 0, 2, 1 };

inline
taka1Board::taka1Board()
{
	;
}

inline
taka1Board::taka1Board(taka1Board &src) :
	mycount(src.mycount), matchcount(src.matchcount), totalcount(totalcount),
	step(src.step), myflag(src.myflag)
{
	memcpy(t, src.t, sizeof(piece_t)*TABLE_SIZE);
}

taka1Board::taka1Board(char *initboard)
{
	int l, d;

	for (l=0; l<9; l++) {
		t[tbl(0, l)] =
		t[tbl(l, 0)] =
		t[tbl(l, 9)] = -1;
	}
	t[tbl(9,9)] = -1;
	mycount = matchcount = 0;
	totalcount = 0;
	step = my_step;
	myflag = 1;
	for (l=0; l<64; l++) {
		t[intable_table[l]] = d = initboard[l];
		if (d == P_MY_COLOR) mycount++;
		else if (d == P_OP_COLOR) matchcount++;
	}
	totalcount = mycount + matchcount;
}

inline int
taka1Board::table(int pos)
{
	return t[pos];
}

inline void
taka1Board::settable(int pos, int d)
{
	t[pos] = d;
}

void
taka1Board::put_reverse(taka1Board &dstboard)
{
	int l;

	dstboard.mycount = matchcount;
	dstboard.matchcount = mycount;
	dstboard.totalcount = mycount + matchcount;
	if (myflag) {
		dstboard.step = step;
		dstboard.myflag = 0;
	}else {
		dstboard.step = step+1;
		dstboard.myflag = 1;
	}
	for (l=0; l<TABLE_SIZE; l++) {
		dstboard.t[l] = rev_table[1+t[l]];
	}
}

inline taka1Board *
taka1Board::create_reverse()
{
	taka1Board *rp = new taka1Board();
	put_reverse(*rp);
	return rp;
}

#if TAKADEBUG
void
taka1Board::dump()
{
	int x, y;
	int my, match;

	cout <<"taka: taka1Board " <<this <<" -> "
	  <<"mycount=" <<(int)mycount
	  <<", matchcount=" <<(int)matchcount
	  <<", step=" <<(int)step
	  <<", myflag=" <<(int)myflag <<endl;
	
	my = match = 0;
	for (y=0; y<=9; y++) {
		cout <<"  " <<y <<":";
		for (x=0; x<9; x++) {
			cout <<setw(3) <<(int)t[y*9+x];
			if (t[y*9+x] == P_MY_COLOR) my++;
			else if (t[y*9+x] == P_OP_COLOR) match++;
		}
		if (y==9) cout <<" ... " << (int)t[y*9+x];
		cout <<endl;
	}
	
	if (my != mycount) cout <<"  mycount mismatch! " <<my <<endl;
	if (match != matchcount) cout <<"  matchcount mismatch! " <<match <<endl;
}
#endif


// class taka1Hand
inline
taka1Hand::taka1Hand() : th(this)
{
	//tp = new Thinker(this);
}

inline
taka1Hand::~taka1Hand()
{
	//if (tp) delete tp;
}

inline Thinker &
taka1Hand::think()
{
	return th;
}


// class taka1Choice
taka1Hand *taka1Choice::handstock = new taka1Hand();

taka1Choice::taka1Choice(taka1Choice &/*src*/)
{
	abort();
}
taka1Choice::taka1Choice() : hc(0)
{
	abort();
}

taka1Choice::~taka1Choice()
{
	int l;

	if (!hc) {
		/* If hc==0, hv[0] keeps valid object to destroy. */
		delete hv[0];
#if TAKADEBUG
		hv[0] = (taka1Hand *)DEADBEEF;
#endif
		return;
	}
	for (l=0; l<hc; l++) {
		delete hv[l];
#if TAKADEBUG
		hv[l] = (taka1Hand *)DEADBEEF;
#endif
	}
}

inline void
taka1Choice::erasehand(int handcount)
{
	delete hv[handcount];
	hv[handcount] = NULL;
}

taka1Choice::taka1Choice(taka1Board &from)
{
	int p;
	int l;
	//taka1Hand *hp;

#if TAKADEBUG
	for (l=0; l<maxhand; l++) hv[l] = (taka1Hand *)DEADBEEF;
#endif
	hc = 0;
	//hp = new taka1Hand();
	for (l=0; l<64; l++) {
		p = intable_table[l];
		if (from.table(p) != P_SPACE) continue;
		if (
		  from.table(p+moves_table[0]) != P_OP_COLOR &&
		  from.table(p+moves_table[1]) != P_OP_COLOR &&
		  from.table(p+moves_table[2]) != P_OP_COLOR &&
		  from.table(p+moves_table[3]) != P_OP_COLOR &&
		  from.table(p+moves_table[4]) != P_OP_COLOR &&
		  from.table(p+moves_table[5]) != P_OP_COLOR &&
		  from.table(p+moves_table[6]) != P_OP_COLOR &&
		  from.table(p+moves_table[7]) != P_OP_COLOR
		) {
			continue;
		}
		if (makeput(handstock, from, p)) {
			hv[hc++] = handstock;
			handstock = new taka1Hand();
		}
	}
	if (hc == 0) {
		/* pass */
		from.put_reverse(handstock->bd);
		/* handstock->bd = from; */
		handstock->getcount = 0;
		handstock->pos = taka1Hand::passed;
		//handstock->tp = new Thinker(handstock);
		hv[0] = handstock;
		handstock = new taka1Hand();
	}else {
		//delete hp;
	}
}

inline int
taka1Choice::number() {
	return hc;
}

inline taka1Hand &
taka1Choice::item(int hnum) {
	return *hv[hnum];
}

int
taka1Choice::makeput(taka1Hand *handp, taka1Board &from, int pos)
{
	int l, p, md;
	int mc, gc;
	getlist_t *gv;

#if TAKADEBUG*0
	cout <<"makeput: " <<pos <<"(" <<tblx(pos) <<"," <<tbly(pos) <<")" <<endl;
#endif
	gc = 0;
	for (mc=0; mc<8; mc++) {
		if (from.table(pos+moves_table[mc]) != P_OP_COLOR) continue;
		md = moves_table[mc];
		gv = handp->getlist + gc;
		*gv = p = pos+md;
		l=1;
		while (from.table(p+=md) == P_OP_COLOR) {
			*(gv + l++) = p;
		}
		if (from.table(p) != P_MY_COLOR) continue;
		gc += l;
	}
	if (gc == 0) {
		return 0;
	}
	
	// update board table
	from.put_reverse(handp->bd);
	//handp->bd = from;
	handp->bd.settable(pos, P_OP_COLOR);
	for (l=0; l<gc; l++) {
		handp->bd.settable(handp->getlist[l], P_OP_COLOR);
	}
	handp->bd.matchcount += 1+gc;
	handp->bd.mycount -= gc;
	handp->bd.totalcount++;
	handp->pos = pos;
	handp->getcount = gc;
	//handp->tp = new Thinker(handp);
	
	return 1;
}

void
taka1Choice::handsort(int usecount)
{
	taka1Hand *hp;

	switch (number()) {
	case 0:
	case 1:
		return;
	case 2:
		if (item(0).think().weight > item(1).think().weight) {
			hp=hv[0]; hv[0]=hv[1]; hv[1]=hp;
		}
		return;
	case 3:
		if (item(0).think().weight > item(1).think().weight) {
			if (item(0).think().weight > item(2).think().weight) {
				if (item(1).think().weight > item(2).think().weight) {
					// 2->1->0
					hp=hv[0]; hv[0]=hv[2]; hv[2]=hp;
				}else {
					// 1->2->0
					hp=hv[0]; hv[0]=hv[1]; hv[1]=hv[2]; hv[2]=hp;
				}
			}else {
				// 1->0->2
				hp=hv[0]; hv[0]=hv[1]; hv[1]=hp;
			}
		}else {
			if (item(1).think().weight > item(2).think().weight) {
				if (item(0).think().weight > item(2).think().weight) {
					// 2->0->1
					hp=hv[0]; hv[0]=hv[2]; hv[2]=hv[1]; hv[1]=hp;
				}else {
					// 0->2->1
					hp=hv[1]; hv[1]=hv[2]; hv[2]=hp;
				}
			}
		}
		return;
	case 4:
		if (item(0).think().weight > item(2).think().weight) {
			hp=hv[0]; hv[0]=hv[2]; hv[2]=hp;
		}
		if (item(1).think().weight > item(3).think().weight) {
			hp=hv[1]; hv[1]=hv[3]; hv[3]=hp;
		}
		if (item(0).think().weight > item(1).think().weight) {
			hp=hv[0]; hv[0]=hv[1]; hv[1]=hp;
		}
		if (item(2).think().weight > item(3).think().weight) {
			hp=hv[2]; hv[2]=hv[3]; hv[3]=hp;
		}
		if (item(1).think().weight > item(2).think().weight) {
			hp=hv[1]; hv[1]=hv[2]; hv[2]=hp;
		}
		return;
	}
	{
		int e, l, i, p, m;
		
		// selective sort
		e = 2;
		if (usecount > e) e = usecount;
		if (e > number()-1) e = number()-1;
		for (l=0; l<e; l++) {
			m = item(l).think().weight;
			p = l;
			for (i=l+1; i<hc; i++) {
				if (item(i).think().weight < m) {
					m = item(i).think().weight;
					p = i;
				}
			}
			if (l != p) {
				hp=hv[l]; hv[l]=hv[p]; hv[p]=hp;
			}
		}
	}
}


// class taka1Think
taka1Think::taka1Think(taka1Think &/*src*/)
{
	abort();
}

taka1Think::taka1Think()
{
	abort();
}

inline
taka1Think::taka1Think(taka1Hand *handp) : handp(handp)
{
#if TAKADEBUG
	weight = 0;
#endif
	;
}

inline
taka1Think::~taka1Think()
{
#if TAKADEBUG
		handp = (taka1Hand *)DEADBEEF;
#endif
	;
}

inline taka1Hand &
taka1Think::hand()
{
	return *handp;
}

inline taka1Board &
taka1Think::board()
{
	return handp->bd;
}


/*----------------------------------------------------------------------*/
/*
 * utility and initialization functions
 */
STATIC void
initialize_tables(void)
{
	int x, y;

	for (y=0; y<8; y++) for (x=0; x<8; x++) {
		intable_table[y*8 + x] = tbl(x+1, y+1);
	}
#if TAKADEBUG*0
	for (int l=0; l<64; l++) {
		if (l%8 == 0) cout <<endl;
		cout <<setw(3) <<intable_table[l];
	}
	cout <<endl;
#endif
#if TAKADEBUG
	cout <<"initialize_tables: sizeof... taka1Hand=" <<sizeof(taka1Hand)
	  <<", taka1Choice=" <<sizeof(taka1Choice)
	  <<", Thinker=" <<sizeof(Thinker) <<"." <<endl;
#endif
	
	if (sizeof(taka1Choice) > 256) {
#if TAKADEBUG
		cout <<"taka: sizeof(taka1Choice)=" <<sizeof(taka1Choice) <<endl;
#endif
		abort();
	}
	if (sizeof(taka1Hand) > 256) {
#if TAKADEBUG
		cout <<"taka: sizeof(taka1Hand)=" <<sizeof(taka1Hand) <<endl;
#endif
		abort();
	}
}


/*
 * entry of thinking function
 */
extern "C" int
taka1(char *board, int status)
{
	int p;

	if (status) {
		if (initp(status)) {
			/* initialization for static variables */
			my_step = 0;
			initialize_tables();
			Thinker::initialize();
//printf("taka: init!(%d)\n", status);
			
			return 0;
		}
		if (passp(status)) {
			/* pass! -  check match's hand */
			my_step++;
			
			return 0;
		}
#if TAKADEBUG
		cout <<"taka .. unknown status." <<endl;
#endif
		abort();
	}
	
	my_step++;
	taka1Board bd(board);
	taka1Choice chfirst(bd);
	if (chfirst.number() <= 1) {
		if (chfirst.number() == 0) {
			/* pass */
			//return -1;
#if TAKADEBUG
			cout <<"taka .. unexpected pass situation." <<endl;
#endif
			abort();
		}
		/* no room to choose */
		p = chfirst.item(0).pos;
		goto thank;
	}
	
	p = Thinker::think(bd, chfirst);

thank:
	return tblresult(p);
}


/*----------------------------------------------------------------------*/

/*
 * Thinking routine #1: 1997/09/26(Fri.)
 */
// static variables
int taka1Think1::weight_table_seed[4][4] = {
	{	60,	1,	14,	10,	},
	{	1,	6,	12,	8,	},
	{	14,	12,	5,	6,	},
	{	10,	8,	6,	6,	},
};
int taka1Think1::weight_table_seed2[4][4] = {
	{	99,	-5,	28,	10,	},
	{	-5,	6,	26,	8,	},
	{	28, 26,	5,	6,	},
	{	10,	8,	6,	6,	},
};
int taka1Think1::weight_table[TABLE_SIZE];

// functions
taka1Think1::taka1Think1(taka1Hand *handp) : taka1Think(handp)
{
	cp = NULL;
	/* bp = NULL; */
}

taka1Think1::~taka1Think1()
{
	if (cp != NULL) {
		delete cp;
#if TAKADEBUG
		cp = (taka1Choice *)DEADBEEF;
#endif
	}
#if 0
	if (bp != NULL) {
		delete bp;
#if TAKADEBUG
		bp = (taka1Board *)DEADBEEF;
#endif
	}
#endif
}

void
taka1Think1::initialize()
{
	int l, x, y;

	for (l=0; l<=9; l++) {
		weight_table[tbl(0, l)] =
		weight_table[tbl(9, l)] =
		weight_table[tbl(l, 0)] =
		weight_table[tbl(l, 9)] = 0;
	}
	for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
		weight_table[tbl(x,   y  )] =
		weight_table[tbl(9-x, y  )] =
		weight_table[tbl(x,   9-y)] =
		weight_table[tbl(9-x, 9-y)] = weight_table_seed2[y-1][x-1];
	}
#if TAKADEBUG*0
	for (l=0; l<91; l++) {
		if (l%9 == 0) cout <<endl;
		cout <<setw(3) <<weight_table[l];
	}
	cout <<endl;
#endif
	
	
}

void
taka1Think1::eval()
{
	int l;
	int v;
	int d;
	int x,y;
	static int weight_mpl[3] = { 0, 1, -1 };
	static int weight_table_a1[4][4] = {
		{	99,	8,	10,	8,	},
		{	8,	6,	7,	8,	},
		{	10,	7,	6,	5,	},
		{	8,	8,	5,	5,	},
	};
	static int weight_table_a2[4][4] = {
		{	200,-30,10,	9,	},
		{	-20,10,	8,	9,	},
		{	28,	24,	5,	7,	},
		{	10,	8,	6,	6,	},
	};
	static int weight_table_a3[4][4] = {
		{	200,-19,13, 11,	},
		{	-19,6,	9,	8,	},
		{	13,	9,	6,	5,	},
		{	11,	8,	5,	5,	},
	};

	if (board().mycount + board().matchcount < 28) {
		v = 0;
		if (board().table(tbl(1,1)) != P_SPACE) {
			for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
				v += weight_table_a1[y-1][x-1] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else if (board().table(tbl(2,1)) != P_SPACE) {
			if (board().table(tbl(1,2)) != P_SPACE) {
				for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
					v += weight_table_a3[y-1][x-1] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}else {
				for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
					v += weight_table_a2[y-1][x-1] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}
		}else if (board().table(tbl(1,2)) != P_SPACE) {
			for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
				v += weight_table_a2[x-1][y-1] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else {
			for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
				v += weight_table[tbl(x, y)] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}
		
		if (board().table(tbl(8,8)) != P_SPACE) {
			for (y=8; y>=5; y--) for (x=8; x>=5; x--) {
				v += weight_table_a1[8-y][8-x] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else if (board().table(tbl(7,8)) != P_SPACE) {
			if (board().table(tbl(8,7)) != P_SPACE) {
				for (y=8; y>=5; y--) for (x=8; x>=5; x--) {
					v += weight_table_a3[8-y][8-x] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}else {
				for (y=8; y>=5; y--) for (x=8; x>=5; x--) {
					v += weight_table_a2[8-y][8-x] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}
		}else if (board().table(tbl(8,7)) != P_SPACE) {
			for (y=8; y>=5; y--) for (x=8; x>=5; x--) {
				v += weight_table_a2[8-x][8-y] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else {
			for (y=8; y>=5; y--) for (x=8; x>=5; x--) {
				v += weight_table[tbl(x, y)] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}
		
		// hidari-shita
		if (board().table(tbl(1,8)) != P_SPACE) {
			for (y=8; y>=5; y--) for (x=1; x<=4; x++) {
				v += weight_table_a1[8-y][x-1] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else if (board().table(tbl(2,8)) != P_SPACE) {
			if (board().table(tbl(1,7)) != P_SPACE) {
				for (y=8; y>=5; y--) for (x=1; x<=4; x++) {
					v += weight_table_a3[8-y][x-1] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}else {
				for (y=8; y>=5; y--) for (x=1; x<=4; x++) {
					v += weight_table_a2[8-y][x-1] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}
		}else if (board().table(tbl(1,7)) != P_SPACE) {
			for (y=8; y>=5; y--) for (x=1; x<=4; x++) {
				v += weight_table_a2[x-1][8-y] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else {
			for (y=8; y>=5; y--) for (x=1; x<=4; x++) {
				v += weight_table[tbl(x, y)] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}
		
		// migi-ue
		if (board().table(tbl(8,1)) != P_SPACE) {
			for (y=1; y<=4; y++) for (x=8; x>=5; x--) {
				v += weight_table_a1[y-1][8-x] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else if (board().table(tbl(7,1)) != P_SPACE) {
			if (board().table(tbl(8,2)) != P_SPACE) {
				for (y=1; y<=4; y++) for (x=8; x>=5; x--) {
					v += weight_table_a3[y-1][8-x] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}else {
				for (y=1; y<=4; y++) for (x=8; x>=5; x--) {
					v += weight_table_a2[y-1][8-x] *
					     weight_mpl[board().table(tbl(x, y))];
				}
			}
		}else if (board().table(tbl(8,2)) != P_SPACE) {
			for (y=1; y<=4; y++) for (x=8; x>=5; x--) {
				v += weight_table_a2[8-x][y-1] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}else {
			for (y=1; y<=4; y++) for (x=8; x>=5; x--) {
				v += weight_table[tbl(x, y)] *
				     weight_mpl[board().table(tbl(x, y))];
			}
		}
		
		 /* board() has the state in complementary color */
		weight = - v;
		return;
	}
	
	if (board().mycount + board().matchcount < 60/*40*/) {
		v = 0;
		for (l=0; l<64; l++) {
			d = board().table(intable_table[l]);
			if (d == P_MY_COLOR) {
				v += weight_table[intable_table[l]];
			}else if (d == P_OP_COLOR) {
				v -= weight_table[intable_table[l]];
			}
		}
		 /* board() has the state in complementary color */
		weight = - v;
		return;
	}
	
	 /* board() has the state in complementary color */
	weight = - (board().mycount - board().matchcount);
	
	return;
}

int
taka1Think1::think(taka1Board &bd, taka1Choice &chfirst)
{
	int l;
	int c;
	int e;
	int lv;
	int max, min, maxpos, minpos;

	if (bd.mycount + bd.matchcount >= 53) {
#if TAKADEBUG
		cout <<"taka: " <<(64 - bd.mycount - bd.matchcount)
		  <<"ꡢɤߤޤ(" <<chfirst.number() <<"̤) "<<flush;
#endif
		min = 9999; minpos = -1;
		for (l=0; l<chfirst.number(); l++) {
#if TAKADEBUG
			cout <<l <<flush;
//cout <<endl;
#endif
			c = - chfirst.item(l).think().finalize(0, min);
#if TAKADEBUG
			cout <<"<" <<c <<"> "<<flush;
#endif
			if (c < min) {
				if (minpos >= 0) chfirst.erasehand(minpos);
				min = c;
				minpos = l;
			}else {
				chfirst.erasehand(l);
			}
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		return chfirst.item(minpos).pos;
	}
	
	
	if (bd.mycount + bd.matchcount >= 48) {
#if TAKADEBUG
		cout <<"taka: " <<(64 - bd.mycount - bd.matchcount)
		  <<"ꡢɤߤޤ(" <<chfirst.number() <<"̤) "<<flush;
#endif
		for (l=0; l<chfirst.number(); l++) {
			chfirst.item(l).think().eval();
		}
		chfirst.handsort();
#if TAKADEBUG*0
		cout <<"sorted.. ";
		for (l=0; l<chfirst.number(); l++) {
			cout <<chfirst.item(l).think().weight <<" ";
		}
		cout <<endl;
#endif
		min = 9999; minpos = -1;
		e = chfirst.number();
		if (e>4) e=4;
		for (l=0; l<e; l++) {
#if TAKADEBUG
			cout <<l <<flush;
#endif
			c = - chfirst.item(l).think().semifinalize(0, min);
#if TAKADEBUG
			cout <<"<" <<c <<"> "<<flush;
#endif
			if (c < min) {
				if (minpos >= 0) chfirst.erasehand(minpos);
				min = c;
				minpos = l;
			}else {
				chfirst.erasehand(l);
			}
			if (l>=1 && min<0) break;
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		return chfirst.item(minpos).pos;
	}
	
	
	if (bd.mycount + bd.matchcount >= 9) {
		lv = 3-1;
	if (bd.mycount + bd.matchcount >= 14) {
		lv = 5-1;
		if (bd.mycount + bd.matchcount >= 28) {
			lv = 6-1;
			if (bd.mycount + bd.matchcount >= 38) {
				lv = 7-1;
				if (bd.mycount + bd.matchcount >= 44) {
					lv = 8-1;
				if (bd.mycount + bd.matchcount >= 47) {
					lv = 9-1;
				}
				}
			}
		}
		}
#if TAKADEBUG
		cout <<"taka: " <<(64 - bd.mycount - bd.matchcount) <<"ꡢ"
		  <<lv+1 <<"ɤߤޤ(" <<chfirst.number() <<"̤)" <<flush;
#endif
		for (l=0; l<chfirst.number(); l++) {
			chfirst.item(l).think().eval();
		}
		chfirst.handsort();
#if TAKADEBUG*1
		cout <<"sorted.. ";
		for (l=0; l<chfirst.number(); l++) {
			cout <<chfirst.item(l).think().weight <<" ";
		}
		cout <<endl;
#endif
		if (bd.mycount + bd.matchcount == 26 || bd.mycount + bd.matchcount >= 16 && bd.mycount + bd.matchcount < 26 && chfirst.item(0).think().weight < -10) {
			int x,y;
			for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
				weight_table[tbl(x,   y  )] =
				weight_table[tbl(9-x, y  )] =
				weight_table[tbl(x,   9-y)] =
				weight_table[tbl(9-x, 9-y)] = weight_table_seed[y-1][x-1];
			}
		}
	
		min = 9999; minpos = -1;
		e = chfirst.number();
		if (e>4) e=4;
		for (l=0; l<e; l++) {
#if TAKADEBUG
			cout <<l <<flush;
#endif
			c = - chfirst.item(l).think().nthforward(0, min, lv);
#if TAKADEBUG
			cout <<"<" <<c <<"> "<<flush;
#endif
			if (c < min) {
				if (minpos >= 0) chfirst.erasehand(minpos);
				min = c;
				minpos = l;
			}else {
				chfirst.erasehand(l);
			}
			if (l>=1 && min<0) break;
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		return chfirst.item(minpos).pos;
	}
	
	
	for (l=0; l<chfirst.number(); l++) {
		chfirst.item(l).think().eval();
	}
	max = -9999; min = 9999; maxpos=minpos=0;
	for (l=0; l<chfirst.number(); l++) {
		if (chfirst.item(l).think().weight > max) {
			max = chfirst.item(l).think().weight;
			maxpos = l;
		}
		if (chfirst.item(l).think().weight < min) {
			min = chfirst.item(l).think().weight;
			minpos = l;
		}
	}
	
	return chfirst.item(minpos).pos;
}

int
taka1Think1::finalize(int passcount, int orgmin)
{
	int l;
	int c, min, minpos;

	if (board().mycount + board().matchcount == 64) {
		return (board().mycount - board().matchcount);
	}
	/* bp = board().create_reverse(); */
	cp = new taka1Choice(board());
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				/* both pass */
				return (board().mycount - board().matchcount);
			}
			return - cp->item(0).think().finalize(1, 9999);
		}
		/* no room to choose */
		return - cp->item(0).think().finalize(0, 9999);
	}
	
	min = 9999; minpos = -1; c = 0;
	for (l=0; l<cp->number(); l++) {
		cp->item(l).think().aheadresult
		  = c
		  = - cp->item(l).think().finalize(0, min);
		if (c < min) {
			if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			cp->erasehand(l);
		}
		if (orgmin <= -min) break;	// alpha-beta
	}

	return min;
}

int
taka1Think1::semifinalize(int passcount, int orgmin)
{
	int l;
	int c, min, minpos;
	int e;

	if (board().mycount + board().matchcount >= 60) return finalize(passcount, orgmin);
#if TAKADEBUG*0
cout <<"semifinalize() level "
  <<(64 - board().mycount - board().matchcount)
  <<" passcount=" <<passcount <<" myflag=" <<(int)board().myflag
  <<" mycount=" <<(int)board().mycount
  <<" matchcount=" <<(int)board().matchcount <<endl;
#endif
	if (board().mycount + board().matchcount == 64) {
//cout <<"  64filled. return " <<(board().mycount - board().matchcount)
//  <<endl;
		return (board().mycount - board().matchcount);
	}
	/*bp = board().create_reverse();*/
	cp = new taka1Choice(board()/* *bp */);
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				// both pass
#if TAKADEBUG*0
cout <<"  bothpass. return " <<(board().mycount - board().matchcount)
  <<endl;
#endif
				return (board().mycount - board().matchcount);
			}
//cout <<"  onepass." <<endl;
			return - cp->item(0).think().semifinalize(1, 9999);
		}
		/* no room to choose */
//cout <<"  no room to choose." <<endl;
		return - cp->item(0).think().semifinalize(0, 9999);
	}
	
#if TAKADEBUG*0
cout <<"  next level. (" <<cp->number() <<")" <<endl;
#endif
	e = cp->number();
	for (l=0; l<cp->number(); l++) {
		cp->item(l).think().eval();
	}
	cp->handsort();
#if 0
	if (e >= 4) {
		e = 4;
		if (board().mycount + board().matchcount > 54) e = 3;
	}
#endif
	if (e >= 3) e = 3;
	min = 9999; minpos = -1; c = 0;
	for (l=0; l<e; l++) {
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" item " <<l <<endl;
#endif
		cp->item(l).think().aheadresult
		  = c
		  = - cp->item(l).think().semifinalize(0, min);
		if (c < min) {
			if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			cp->erasehand(l);
		}
		if (orgmin <= -min) break;	// alpha-beta
		/*
		if (l==1 && min<0 && (board().mycount + board().matchcount <= 54))
			break;
		*/
	}
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" exit -> " <<minpos <<"<" <<min <<">" <<endl;
//cout <<"     ";
//for (l=0; l<cp->number(); l++) cout <<l <<"<" <<cp->item(l).think().aheadresult <<"> ";
cout <<endl;
#endif

	return min;
}

int
taka1Think1::nthforward(int passcount, int orgmin, int depth)
{
	int l;
	int c, min, minpos;
	int e;

	if (board().mycount + board().matchcount == 64) {
		return (board().mycount - board().matchcount)*1000;
	}
	if (depth == 0) {
		return -weight;
	}
	/* bp = board().create_reverse(); */
	cp = new taka1Choice(board());
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				// both pass
				return (board().mycount - board().matchcount)*100;
			}
//cout <<"  onepass." <<endl;
			return - cp->item(0).think().nthforward(1, 9999, depth-1);
		}
		/* no room to choose */
//cout <<"  no room to choose." <<endl;
		return - cp->item(0).think().nthforward(0, 9999, depth-1);
	}
	
#if TAKADEBUG*0
cout <<"  next level. (" <<cp->number() <<")" <<endl;
#endif
	e = cp->number();
	for (l=0; l<cp->number(); l++) {
		cp->item(l).think().eval();
	}
	cp->handsort();
#if 1
	if (e >= 4) {
		e = 4;
		if (board().mycount + board().matchcount > 54) e = 3;
	}
#else
	if (e >= 3) e = 3;
#endif
	min = 9999; minpos = -1; c = 0;
	for (l=0; l<e; l++) {
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" item " <<l <<endl;
#endif
		cp->item(l).think().aheadresult
		  = c
		  = - cp->item(l).think().nthforward(0, min, depth-1);
		if (c < min) {
			if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			cp->erasehand(l);
		}
		if (orgmin <= -min) break;	// alpha-beta
		if (depth < 4 && l==2 && min<0 && (board().mycount + board().matchcount <= 54))
			break;
	}
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" exit -> " <<minpos <<"<" <<min <<">" <<endl;
//cout <<"     ";
//for (l=0; l<cp->number(); l++) cout <<l <<"<" <<cp->item(l).think().aheadresult <<"> ";
cout <<endl;
#endif

	return min;
}

/*----------------------------------------------------------------------*/
